import json
import random
from typing import NewType

from jinja2 import Template

from genie_pkg.generators import *
import string

def random_string_list(list_size, item_length):
    l = []
    for _ in range(0, list_size):
        l.insert(0, ''.join([random.choice(string.ascii_letters)
                             for _ in range(0, item_length)]))
    # json dumps is to get python list as json compatible string.
    return json.dumps(l)


def random_integer_list(list_size, start, max):
    l = []
    for _ in range(0, list_size):
        l.insert(0, random_integer(start, max))
    return l


def random_choice_of(choices) -> str:
    return str(one_of(choices))


def random_email_id(width, domain) -> str:
    return generate_email_id(width, domain)


def random_ipv4() -> str:
    return generate_ip()


def date_with_format(format_string='%Y/%m/%d', delta_days=0):
    return random_date_from_today(format_string, delta_days)


def random_geo(center=(-37.814, 144.963,), radius=10000, accuracy=3):
    x0, y0 = random_geo_coords(center, radius, accuracy)
    return json.dumps({"latitude": x0, "longitude": y0})


def random_text():
    return random_wonderland_text()

JinjaTemplate = NewType('JinjaTemplate', Template)


def _add_template_functions(template: JinjaTemplate) -> JinjaTemplate:
    template.globals['random_integer'] = random_integer
    template.globals['random_float'] = random_float
    template.globals['random_string'] = random_string
    template.globals['random_string_with_special_chars'] = random_string_with_special_chars
    template.globals['random_string_list'] = random_string_list
    template.globals['now_epoch'] = now_epoch
    template.globals['random_bool'] = random_bool
    template.globals['random_integer_list'] = random_integer_list
    template.globals['guid'] = guid
    template.globals['random_choice_of'] = random_choice_of
    template.globals['date_with_format'] = date_with_format
    template.globals['random_email_id'] = random_email_id
    template.globals['random_ipv4'] = random_ipv4
    template.globals['random_geo'] = random_geo
    template.globals['random_mastercard_number'] = random_mastercard_number
    template.globals['random_visacard_number'] = random_visacard_number
    template.globals['random_text'] = random_text
    return template


def generate(template_string) -> str:
    '''
        Generate json data for the provided specification

        Args:
            template_string (str): Jinja template string

        Returns:
            data (str): Jinja template rendered.
    '''

    t = Template(template_string)
    return _add_template_functions(t).render()


def generate_with_custom_template_function(template: JinjaTemplate) -> str:
    '''
        Generate json data for the provided specification

        Args:
            template (Template): Jinja template

        Returns:
            data (str): Jinja template rendered.
    '''

    return _add_template_functions(template).render()



import pandas as pd
import pandas_flavor as pf
from lark import Lark, Tree
from pkg_resources import resource_string

@pf.register_dataframe_accessor('dqc')
class QualityChecker(object):
    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj

    @staticmethod
    def _validate(obj):
        if obj.dropna().empty:
            raise AttributeError("DataFrame is empty!!! What quality do you expect from emptiness")

    def _apply_check(self, check):
        try:
            c = check[0]
            if c.data == "has_size":
                quantity = int(c.children[0])
                return c.data, self._obj.shape[0] == quantity
            elif c.data == "has_columns":
                column_names = []
                for ct in c.children[0].children:
                    name = ct.value
                    column_names.insert(0, name.replace("'", ""))
                return c.data, len(set(self._obj.columns).intersection(column_names)) == len(column_names)
            elif c.data == "is_unique":
                column_name = c.children[0]
                return c.data, self._obj[column_name].is_unique
            else:
                return c.data, False
        except KeyError as ke:
            return "Key error: {0} for {1}".format(c.data, c.children[0]), False

    def _apply_predicates(self, predicates):
        results = []
        for predicate in predicates:
            if predicate.data == "code_block":
                for check in predicate.children:
                    v = self._apply_check(check.children)
                    results.insert(0, v)
            else:
                raise SyntaxError('Unknown instruction: %s' % preds.data)
        return results


    def _run(self, pt: Tree):
        for instruction in pt.children:
            if instruction.data == "predicates":
                return self._apply_predicates(instruction.children)
            else:
                raise SyntaxError('Unknown instruction: %s' % instruction.data)


    @staticmethod
    def _parse(check_spec) -> Tree:
        p = Lark(resource_string(__name__, 'data/grammar.g').decode('utf-8'))
        return p.parse(check_spec)


    def run(self, check_spec):
        ast = self._parse(check_spec)
        return self._run(ast)






from pkg_resources import resource_string
import json
from genie_pkg.generators import one_of, random_geo_coords

class Australia(object):
    def __init__(self):
        self.data = json.loads(resource_string(__name__, 'data/oz_postcodes.json'))

    
    def get_random_state(self) -> str:
        return one_of(list(self.data.keys()))

    def get_random_city_postcode(self, state) -> (str, str):
        state_data = self.data[state]
        post_codes = list(state_data.keys())      
        pc = one_of(post_codes)
        locality = state_data[pc]["localities"]
        return (locality, pc)

    def get_city(self, state, postcode):
        localities = self.data[state][postcode]["localities"]
        return one_of(localities)
        
    def get_random_geo_coordinate(self, state, postcode):
        center = self.data[state][postcode].get("center", {})
        if center:
            c = (float(center["latitude"]), float(center["longitude"]),)
            return random_geo_coords(center=c)
        else:
            raise Exception("Postcode {0} in state {1} does not have geo center".format(postcode, state))